#include <cstdio>
#include <cstdlib>
#include <algorithm>
#include <numeric> 
#include <random>
#include <vector>
#include <string>

#undef NDEBUG
#include <cassert>
#include <ctime>
#include <cstdint>

#include "simTree.h"
#include "search.h"

#include "utils.h"
#include "statistics.h"

#ifdef _MSC_BUILD
#pragma warning(disable:4996)
#endif



#define SEEDOFFSETroot 4656624
#define SEEDOFFSETBLs  991377



// GLOBALS
VariableSimTreeConfig variabletreeConfig;
SimTreeConfig currenttreeConfig;
ESPRITConfig espritConfig;
PUCTConfig puctConfig;
FILE* debugInfo_fp;


static void readConfig(const char * fpath) {
	FILE *fp = openfileForReading(fpath);
	fscanfVariableSimtreeConfig(fp);
	fscanfESPRITConfig(fp);
	fscanfPUCTConfig(fp);
	fclose(fp);

}


// choose a random value from vec
static double randomElement(std::vector<double> vec, std::default_random_engine & prng) {
	std::uniform_int_distribution<int> unif(0, (int)vec.size() - 1);
	return vec[unif(prng)];
}


// sets parameters for tree (prng seeds, choices from variable config)
static void setCurrentTreeConfig(int treeIdx) {
	currenttreeConfig.treeIndex = treeIdx;
	currenttreeConfig.rootSeed = variabletreeConfig.rootSeed + treeIdx;
	currenttreeConfig.BLsSeed = variabletreeConfig.BLsSeed + treeIdx;

	std::default_random_engine prng;
	prng.seed(currenttreeConfig.rootSeed);

	currenttreeConfig.nMovesOfRootsParent = randomElement(variabletreeConfig.nMovesOfRootsParentSet, prng);
	currenttreeConfig.nMovesScale = randomElement(variabletreeConfig.nMovesScaleSet, prng);
	currenttreeConfig.nMovesStdDev = randomElement(variabletreeConfig.nMovesStdDevSet, prng);

	currenttreeConfig.trueRootEval = randomElement(variabletreeConfig.trueRootEvalSet, prng);
	currenttreeConfig.trueEvalspread = randomElement(variabletreeConfig.trueEvalspreadSet, prng);

	currenttreeConfig.estEvaldf = randomElement(variabletreeConfig.estEvaldfSet, prng);
	currenttreeConfig.estEvalscalePar = randomElement(variabletreeConfig.estEvalscaleParSet, prng);
	currenttreeConfig.estEvalscaleCh = randomElement(variabletreeConfig.estEvalscaleChSet, prng);

	currenttreeConfig.policyEvaldf = randomElement(variabletreeConfig.policyEvaldfSet, prng);
	currenttreeConfig.policyEvalscaleCh = randomElement(variabletreeConfig.policyEvalscaleChSet, prng);
	currenttreeConfig.policyEvalscaleEE = randomElement(variabletreeConfig.policyEvalscaleEESet, prng);
	currenttreeConfig.policySoftmaxScale = randomElement(variabletreeConfig.policySoftmaxScaleSet, prng);
}



// runs a search for one method on range of trees i0 to i1.
// outputs to screen and three files
// 1 for first tree only, outputs whole tree as text file
// 2 it collects data from each search tree and outputs it to raw results file
// 3 it outputs a brief summary of same data, with config details
static void runTreeSearches(int i0, int i1, SearchMethod method, 
	                        const char *treesfpath, const char *summaryfpath, const char *rawresultsfpath) {
	const char *methodtxt = 0;
	switch (method) {
	case SearchMethod::ESPRIT: methodtxt = "ESPRIT"; break;
	case SearchMethod::PUCT: methodtxt = "PUCT"; break;
	}
	printf("Doing trees %d to %d for %s\n", i0, i1, methodtxt);

	FILE *summfp = openFileForWriting(summaryfpath);
	FILE *rawfp = openFileForWriting(rawresultsfpath);
	fprintfVariableSimtreeConfig(summfp);


	// variable simulation params (different for each tree)
	std::vector<int> nMovesOfRootsParent(i1 - i0);
	std::vector<int> nMovesScales(i1 - i0);
	std::vector<double> trueRootEvals(i1 - i0);
	std::vector<double> trueEvalspreads(i1 - i0);
	std::vector<double> estEvaldfs(i1 - i0);
	std::vector<double> estEvalscalePars(i1 - i0);
	std::vector<double> estEvalscaleChs(i1 - i0);
	std::vector<double> policyEvaldfs(i1 - i0);
	std::vector<double> policyEvalscaleChs(i1 - i0);
	std::vector<double> policyEvalscaleEEs(i1 - i0);

	std::vector<int> chosenMoveRanks(variabletreeConfig.nMovesMaximum + 1);
	std::vector<double> evalMeans(i1 - i0);
	std::vector<double> evalMinimaxes(i1 - i0);
	std::vector<double> evalMeanErrors(i1 - i0);
	std::vector<double> evalMinimaxErrors(i1 - i0);
	std::vector<double> regrets(i1-i0);
	switch (method) {
	case SearchMethod::ESPRIT:
		fprintfESPRITConfig(summfp);
		break;
	case SearchMethod::PUCT:
		fprintfPUCTConfig(summfp);
		break;
	}
	for (int i = i0; i < i1; i++) {
		if (i != i0) { // only output first tree
			treesfpath = 0; 
		}

		setCurrentTreeConfig(i);
		nMovesOfRootsParent[i - i0] = (int)currenttreeConfig.nMovesOfRootsParent;
		nMovesScales[i - i0] = (int)currenttreeConfig.nMovesScale;
		trueRootEvals[i - i0] = currenttreeConfig.trueRootEval;
		trueEvalspreads[i - i0] = currenttreeConfig.trueEvalspread;
		estEvaldfs[i - i0] = currenttreeConfig.estEvaldf;
		estEvalscalePars[i - i0] = currenttreeConfig.estEvalscalePar;
		estEvalscaleChs[i - i0] = currenttreeConfig.estEvalscaleCh;
		policyEvaldfs[i - i0] = currenttreeConfig.policyEvaldf;
		policyEvalscaleChs[i - i0] = currenttreeConfig.policyEvalscaleCh;
		policyEvalscaleEEs[i - i0] = currenttreeConfig.policyEvalscaleEE;

		SearchResults res = runSearch(treesfpath, method);
		chosenMoveRanks[res.choice] ++;
		assert(res.choice >= 0);
		if (res.choice < 10) {
			printf("%d", res.choice);
		} else {
			printf(".");
		}
		if ((i - i0) % 100 == 99) {
			printf("\n");
		}
		trueRootEvals[i - i0] = res.trueRootEval;
		evalMeans[i - i0] = res.evalmean;
		evalMinimaxes[i - i0] = res.evalminimax;
		evalMeanErrors[i - i0] = res.evalmean - res.trueRootEval;
		evalMinimaxErrors[i - i0] = res.evalminimax - res.trueRootEval;
		regrets[i - i0] += (res.true1plyEvals[res.choice] - res.true1plyEvals[0]);
	}
	fprintf(rawfp, "tree nMovesOfRootsParent nMovesScale ");
	fprintf(rawfp, "trueRootEval trueEvalspread estEvaldf estEvalscalePar estEvalscaleCh ");
	fprintf(rawfp, "policyEvaldf policyEvalscaleCh policyEvalscaleEE ");
	fprintf(rawfp, "evalMeans  evalMinimaxes  evalMeanErrors  evalMinimaxErrors  regrets\n");
	for (int i = i0; i < i1; i++) {
		fprintf(rawfp, "%d           %d             %d       ", i, nMovesOfRootsParent[i - i0], nMovesScales[i - i0]);
		fprintf(rawfp, " %12f %12f %12f   %12f  %12f  ", trueRootEvals[i - i0], trueEvalspreads[i - i0], estEvaldfs[i - i0], estEvalscalePars[i - i0], estEvalscaleChs[i - i0]);
		fprintf(rawfp, " %12f     %12f   %12f", policyEvaldfs[i - i0], policyEvalscaleChs[i - i0], policyEvalscaleEEs[i - i0]);
		fprintf(rawfp, " %12f   %12f   %12f   %12f  %12f\n", evalMeans[i - i0], evalMinimaxes[i - i0], evalMeanErrors[i - i0], evalMinimaxErrors[i - i0], regrets[i - i0]);
	}
	fclose(rawfp);

	double meanRegret = meanOfVector(regrets);
	double stddevRegret = stddevOfVector(regrets, meanRegret);
	double q50 = quantileOfVector(regrets, 0.5);
	double q75 = quantileOfVector(regrets, 0.75);
	double q95 = quantileOfVector(regrets, 0.95);
	double q99 = quantileOfVector(regrets, 0.99);
	double rmseEvalMean = rmsDiffOfVectors(evalMeans, trueRootEvals);
	double rmseEvalMinimax = rmsDiffOfVectors(evalMinimaxes, trueRootEvals);
	double evalMeanErrorMean = meanOfVector(evalMeanErrors);
	double evalMeanErrorStdDev = stddevOfVector(evalMeanErrors, evalMeanErrorMean);
	double evalMinimaxErrorMean = meanOfVector(evalMinimaxErrors);
	double evalMinimaxErrorStdDev = stddevOfVector(evalMinimaxErrors, evalMinimaxErrorMean);

	fprintf(summfp, "\nChosen move counts: ");
	int tot = 0;
	for (int i = 0; i < 10; i++) {
		tot += chosenMoveRanks[i];
		fprintf(summfp, "%d ", chosenMoveRanks[i]);
		if (i == 4) {
			fprintf(summfp, "  ");
		}
	}
	fprintf(summfp, " (%d others)\n", i1 - i0 - tot);
	double meanRegretSD = stddevRegret / sqrt(regrets.size());
	fprintf(summfp, "Mean regret %f +- %8.6f [%f %f %f %f] \n", 
		meanRegret, meanRegretSD, q50, q75, q95, q99);
	fprintf(summfp, "EvalMean:    rmse %f ErrorMean %f ErrorStdDev %f\n", 
		rmseEvalMean, evalMeanErrorMean, evalMeanErrorStdDev);
	fprintf(summfp, "EvalMinimax: rmse %f ErrorMean %f ErrorStdDev %f\n", 
		rmseEvalMinimax, evalMinimaxErrorMean, evalMinimaxErrorStdDev);
	fprintf(summfp, "The sd for regret is for mean regret over all trees\n");
	fprintf(summfp, "For the estimated root evals via mean and minimax, \n");
	fprintf(summfp, "The RMSE compares estimates to true evals, and\n");
	fprintf(summfp, "the ErrorMean and ErrorStdDev split this into bias and variance.\n");
	fclose(summfp);
	printf(" Done\n");
}







int main(int argc, char **argv) {
	const char *dpath = "D:\\BBB\\ProgOutput\\ESPRITvsPUCT\\2021Feb27\\testconfig";
	if (argc > 1) {
		assert(argc == 2);
		dpath = argv[1];
	}
	char configfpath[300];
	char ESPRITrawresultsfpath[300];
	char PUCTrawresultsfpath[300];
	char ESPRITsummaryfpath[300];
	char PUCTsummaryfpath[300];
	char ESPRITtreesfpath[300];
	char PUCTtreesfpath[300];
	char debugInfofpath[300];
	sprintf(configfpath, "%s\\config.txt", dpath);
	sprintf(ESPRITrawresultsfpath, "%s\\ESPRITrawresults.txt", dpath);
	sprintf(PUCTrawresultsfpath, "%s\\PUCTrawresults.txt", dpath);
	sprintf(ESPRITsummaryfpath, "%s\\ESPRITsummary.txt", dpath);
	sprintf(PUCTsummaryfpath, "%s\\PUCTsummary.txt", dpath);
	sprintf(ESPRITtreesfpath, "%s\\ESPRITtrees.txt", dpath);
	sprintf(PUCTtreesfpath, "%s\\PUCTtrees.txt", dpath);
	sprintf(debugInfofpath, "%s\\debugInfo.txt", dpath);
	readConfig(configfpath);

	debugInfo_fp = openFileForWriting(debugInfofpath); 

	double start = (double)clock() / (double)CLOCKS_PER_SEC;
	if (espritConfig.doESPRIT) {
		runTreeSearches(variabletreeConfig.firstTree, variabletreeConfig.endtrees, SearchMethod::ESPRIT,
			ESPRITtreesfpath, ESPRITsummaryfpath, ESPRITrawresultsfpath);
	}
	double espritDone = (double)clock() / (double)CLOCKS_PER_SEC;

	printf("Time %8.3f\n", espritDone - start);
	if (puctConfig.doPUCT) {
		runTreeSearches(variabletreeConfig.firstTree, variabletreeConfig.endtrees, SearchMethod::PUCT, 
			PUCTtreesfpath, PUCTsummaryfpath, PUCTrawresultsfpath);
	}
	double puctDone = (double)clock() / (double)CLOCKS_PER_SEC;
	printf("Time %8.3f\n", puctDone - espritDone);

	fclose(debugInfo_fp);
	getchar();

	return 0;
}

