#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <algorithm>
#include <numeric> 
#include <random>
#include <vector>
#include <set>

#undef NDEBUG
#include <cassert>
#include <ctime>
#include <cstdint>

#include "simTree.h"
#include "search.h"

#include "utils.h"
#include "statistics.h"



#ifdef _MSC_BUILD
#pragma warning(disable:4996)
#endif

static void search(PPTree & tree, int nPlayouts, SearchMethod searchMethod);




/*********************************************************************************************************/
/********************************************    extern **************************************************/
/*********************************************************************************************************/


SearchResults runSearch(const char *treefpath, SearchMethod searchMethod) {
	assert(espritConfig.nPlayoutsESPRIT == puctConfig.nPlayoutsPUCT);
	int nPlayouts = espritConfig.nPlayoutsESPRIT;
	PPTree tree;
	search(tree, nPlayouts, searchMethod);
	subtreeBackupEvalsRPT(tree, 0); // for final means, minimax
	if (treefpath) {
		writeTree(treefpath, tree);
	}
	return searchResultsFromTree(tree, false);
}


// Bud eval based on parent with setback
static double budEvaluation(double parentEval) {
	double y = exp(espritConfig.setback);
	return (1.0 - parentEval) / (1.0 - parentEval + y * parentEval);
}


// Mixing weight when combining policy-from-evals with initial policy
static double effSampSize(double n) {
	return pow(n, espritConfig.ESSexponent);
}


// Heart of the method. 
// Calculate a policy from evals: v_m, n_m to q_m
// regularize with static policy: p_m, q_m to r_m
// score is r_m / (1 + n_m)
static std::vector<double> calcESPRITscores(double parentEstEval, int nMoves, int nEvalsBelow, std::vector<double> staticPol, std::vector<double> evals, std::vector<int> numEvals) {
	// 1. if no children, return static policy
	if (nEvalsBelow == 0) {
		return staticPol;
	}
	// 2. evals[m] are 0 when numEvals[m]. Put in nonzero value if indicated by espritConfig.setback 
	if (espritConfig.setback < 40.0) {
		double budEval = budEvaluation(parentEstEval); // could combine with other evals for better parent eval 
		for (int mov = 0; mov < nMoves; mov++) {
			if (numEvals[mov] == 0) {
				evals[mov] = budEval;
			}
		}
	}
	// 3. find f(max(v_j)-v_m; gamma, sigma) and normalize 
	double S = espritConfig.evalsToPolicyScale;
	double C = espritConfig.evalsToPolicyC;
	double vvv = 0.0;
	for (int mov = 0; mov < nMoves; mov++) {
		vvv += effSampSize(numEvals[mov]);
	}
	vvv /= nMoves;
	double maxEval = maxOfVector(evals);
	double sd = 1.0 / sqrt(vvv);
	std::vector<double> policyFromEvals(nMoves);
	for (int mov = 0; mov < nMoves; mov++) {
		double x = maxEval - evals[mov];
		x /= sd;
		x /= S;
		policyFromEvals[mov] = pow(1.0 + x, C);
	}
	normaliseProbabilities(policyFromEvals);
	// 4. regularise the eval-based policy with the static policy and normalize again
	for (int mov = 0; mov < (int)policyFromEvals.size(); mov++) {
		if (numEvals[mov] > 0) {
			double ess = effSampSize(numEvals[mov]);
			policyFromEvals[mov] *= ess;
			policyFromEvals[mov] += espritConfig.initialPolicyMix * staticPol[mov];
			policyFromEvals[mov] /= (espritConfig.initialPolicyMix + ess);
		}
		else {
			policyFromEvals[mov] = staticPol[mov];
		}
		assert(!std::isnan(policyFromEvals[mov]));
	}
	normaliseProbabilities(policyFromEvals);
	// 5. Turn into scores
	for (int mov = 0; mov < (int)policyFromEvals.size(); mov++) {
		policyFromEvals[mov] /= (numEvals[mov] + 1.0);
	}
	return policyFromEvals;
}



static std::vector<double> calcPUCTscores(double parentEstEval, int nMoves, int nEvalsBelow, std::vector<double> staticPol, std::vector<double> evals, std::vector<int> numEvals) {
	std::vector<double> scores(nMoves);
	for (int mov = 0; mov < nMoves; mov++) {
		double wt = sqrt(1.0 + nEvalsBelow) / (1.0 + numEvals[mov]);
		double C = puctConfig.puctConstant;
		switch (puctConfig.curve) {
		case Flat: break;
		case LogLikeA0: C += log((numEvals[mov] + puctConfig.puctParamB + 1.0) / puctConfig.puctParamB); break;
		default: assert(0); break;
		}
		scores[mov] = evals[mov] + C * wt * staticPol[mov];
	}
	return scores;
}




static void debugPrintVector(const char *name, std::vector<double> vec) {
	fprintf(debugInfo_fp, "%20s ", name);
	for (int i = 0; i < vec.size(); i++) {
		fprintf(debugInfo_fp, "%8f ", vec[i]);
	}
	fprintf(debugInfo_fp, "\n");
}


static void debugPrintVector(const char *name, std::vector<int> vec) {
	fprintf(debugInfo_fp, "%20s ", name);
	for (int i = 0; i < vec.size(); i++) {
		fprintf(debugInfo_fp, "%8d ", vec[i]);
	}
	fprintf(debugInfo_fp, "\n");
}


static void debugPrintChoiceInfo(const PPTree & tree, int idx, SearchMethod searchMethod,
	double parentEstEval, int nMoves, int nEvalsBelow, 
	std::vector<double> staticPol, std::vector<double> evals, std::vector<int> numEvals, 
	std::vector<double> scores, int mov) {
	if (variabletreeConfig.debugLevel >= 3) {
		const char * smtxt = 0;
		if (searchMethod == ESPRIT) {
			smtxt = "ESPRIT";
		}
		if (searchMethod == PUCT) {
			smtxt = "PUCT";
		}
		assert(smtxt != 0);
		fprintf(debugInfo_fp, "At node idx %d with method %s. Own estEval %f nMoves %d int nEvalsBelow %d. Choice is move %d\n", idx, smtxt, parentEstEval, nMoves, nEvalsBelow, mov);
		debugPrintVector("staticPol", staticPol);
		debugPrintVector("evals", evals);
		debugPrintVector("numEvals", numEvals);
		debugPrintVector("scores", scores);
		if (!tree.nodes[idx].BLCs[mov].isChild) {
			fprintf(debugInfo_fp, "        Move is unexpanded.\n\n");
		}

	}
}




static int bestBLC(const PPTree & tree, int idx, SearchMethod searchMethod) {
	assert(tree.nodes[idx].nEvals > 0);
	int nMoves = (int)tree.nodes[idx].BLCs.size();
	std::vector<double> staticPol(nMoves);
	std::vector<double> evals(nMoves);
	std::vector<int> numEvals(nMoves);
	int nEvalsBelow = collectEvalsAndNumEvals(tree, idx, staticPol, evals, numEvals);
	double parentEstEval = tree.nodes[idx].estEval;
	std::vector<double> scores;
	switch (searchMethod) {
	case ESPRIT: scores = calcESPRITscores(parentEstEval, nMoves, nEvalsBelow, staticPol, evals, numEvals); break;
	case PUCT: scores = calcPUCTscores(parentEstEval, nMoves, nEvalsBelow, staticPol, evals, numEvals); break;
	default: assert(0);
	}
	int mov = argmaxOfVector(scores);
	if (variabletreeConfig.debugLevel >= 3) {

		debugPrintChoiceInfo(tree, idx, searchMethod, parentEstEval, nMoves, nEvalsBelow, staticPol, evals, numEvals, scores, mov);
	}
	return mov;
}


static LeafIndex selectBud(const PPTree & tree, int idx, SearchMethod searchMethod) {
	int mov = bestBLC(tree, idx, searchMethod);
	if (tree.nodes[idx].BLCs[mov].isChild) {
		return selectBud(tree, tree.nodes[idx].BLCs[mov].childIdx, searchMethod);
	}
	LeafIndex lfi;
	lfi.idx = idx;
	lfi.mov = mov;
	return lfi;
}



static void search(PPTree & tree, int nPlayouts, SearchMethod searchMethod) {
	makeRoot(tree);
	for (int i = 0; i < nPlayouts; i++) {
		LeafIndex lfi = selectBud(tree, 0, searchMethod);
		budToChild(tree, lfi);
		doEvalsBackToRoot(tree, (int)tree.nodes.size() - 1);
	}
}




