#include <cstdio>
#include <cstdlib>
#include <algorithm>
#include <numeric> 
#include <random>
#include <vector>
#include <string>

#undef NDEBUG
#include <cassert>
#include <ctime>
#include <cstdint>

#include "simTree.h"

#include "utils.h"
#include "statistics.h"


#ifdef _MSC_BUILD
#pragma warning(disable:4996)
#endif

typedef struct {
	double meanDepth;
	double stddevDepth;
	double meanBranchingFactor;
	double stddevBranchingFactor;
	double meanDevEstEval;
	double stddevDevEstEval;
} TreeStatistics;



/***********************************************************************************************/
/*****************************       tree statistics      **************************************/
/***********************************************************************************************/


static int nChildren(const PPNode & node) {
	int nch = 0;
	for (BudLeafChild leaf : node.BLCs) {
		if (leaf.isChild) {
			nch++;
		}
	}
	return nch;
}


static int nNonTipChildren(const PPTree & tree, int idx) {
	const PPNode & node = tree.nodes[idx];
	int nntch = 0;
	bool isTip = true;
	for (BudLeafChild blc : node.BLCs) {
		if (blc.isChild) {
			isTip = false;
		}
	}
	if (!isTip) {
		nntch = nChildren(node);
	}
	return nntch;
}


static TreeStatistics collectTreeStatistics(const PPTree & tree) {
	TreeStatistics ts;
	std::vector<double> evalDeviations(tree.nodes.size());
	for (int i = 0; i < (int)tree.nodes.size(); i++) {
		const PPNode & node = tree.nodes[i];
		evalDeviations[i] = node.estEval - node.trueEval;
	}
	ts.meanDevEstEval = meanOfVector(evalDeviations);
	ts.stddevDevEstEval = stddevOfVector(evalDeviations, ts.meanDevEstEval);

	std::vector<double> depths(tree.nodes.size());
	for (int i = 0; i < (int)tree.nodes.size(); i++) {
		const PPNode & node = tree.nodes[i];
		depths[i] = (int)node.moveSeq.size();
	}
	ts.meanDepth = meanOfVector(depths);
	ts.stddevDepth = stddevOfVector(depths, ts.meanDepth);

	std::vector<double> branchingF(tree.nodes.size());
	for (int i = 0; i < (int)tree.nodes.size(); i++) {
		branchingF[i] = nNonTipChildren(tree, i);
	}
	ts.meanBranchingFactor = meanOfVector(branchingF);
	ts.stddevBranchingFactor = stddevOfVector(depths, ts.meanBranchingFactor);

	return ts;
}



static void fprintfTreeStats(FILE * fp, TreeStatistics ts) {
	fprintf(fp, "DevEstEval %f +- %f\n", ts.meanDevEstEval, ts.stddevDevEstEval); 
	fprintf(fp, "Depth %f +- %f\n", ts.meanDepth, ts.stddevDepth); 
	// ts.meanBranchingFactor, ts.stddevBranchingFactor is bugged 
}




/***********************************************************************************************/
/*****************************    writing out the tree    **************************************/
/***********************************************************************************************/


static void fprintfPPLeaf(FILE *fp, const BudLeafChild & blc) {
	fprintf(fp, "%6.4f  %6.4f  %6.4f  %6.4f", blc.trueEval, blc.estEval, blc.policy, blc.recurPolicy);
}


static std::vector<size_t> policyRanks(const std::vector<BudLeafChild> &v) {
	std::vector<size_t> idxs(v.size());
	iota(idxs.begin(), idxs.end(), 0);
	sort(idxs.begin(), idxs.end(),
		[&v](size_t i1, size_t i2) { return v[i1].policy > v[i2].policy; } );
	std::vector<size_t> ranks(idxs.size());
	for (int i = 0; i < (int)idxs.size(); i++) {
		size_t pi = idxs[i];
		ranks[pi] = i;
	}
	return ranks;
}


static void fprintfMoveID(FILE *fp, std::vector<int> m) {
	for (int i = 0; i < 12; i++) {
		if (i < (int)m.size()) {
			fprintf(fp, "%2d,", m[i]);
		} else {
			fprintf(fp, "   ");
		}
	}
}


static void fprintfPPNode(FILE *fp, const PPNode & node, int idx) {
	fprintf(fp, "Node %-5d ",idx);
	fprintfMoveID(fp, node.moveSeq);
	fprintf(fp, "\ntrueEval %6.4f, estEval %6.4f, recurSum %6.4f, recurMean %6.4f, recurMaxNeg %6.4f, realProb %6.4f, nEvals %d idx %d children ",
		   node.trueEval, node.estEval, node.recurSum, node.recurSum/node.nEvals, node.recurMaxNeg, node.realProb, node.nEvals, node.parent);
	for (BudLeafChild blc : node.BLCs) {
		if (blc.isChild) {
			fprintf(fp, " %d", blc.childIdx);
		}
	}
	fprintf(fp, "\n  mov trueEval estEval policy recurPol polRank\n");
	std::vector<size_t> ranks = policyRanks(node.BLCs);
	for (int mov = 0; mov < (int)node.BLCs.size(); mov++) {
		fprintf(fp, "   %2d ", mov);
		fprintfPPLeaf(fp, node.BLCs.at(mov));
		fprintf(fp, " %d\n", (int)ranks.at(mov));
	}
	fprintf(fp, "\n\n");
}



static void fprintfBriefPPNode(FILE *fp, const PPNode & node, int idx) {
	fprintf(fp, "%5d  ",idx);
	fprintfMoveID(fp, node.moveSeq);
	fprintf(fp, " %6.4f   %6.4f  %10.4f / %5d = %6.4f   %3d         %d  ",
		node.trueEval, node.estEval, node.recurSum, node.nEvals, node.recurSum / node.nEvals, node.parent, node.rootplayer);
	int nBuds = 0;
	int nLeaves = 0;
	int nChildren = 0;
	for (BudLeafChild blc : node.BLCs) {
		if (isBud(blc)) {
			nBuds++;
		}
		if (isLeaf(blc)) {
			nLeaves++;
		}
		if (isChild(blc)) {
			nChildren++;
		}
	}
	fprintf(fp, " %dM = %dB + %dL + %dC\n", (int)node.BLCs.size(), nBuds, nLeaves, nChildren);
}


// Prints a lot per node, including BLCs.
static void fprintfPPTree(FILE *fp, const PPTree & tree) {
	fprintf(fp, "Node   seed                                 trueEval  estEval  recurSum/nEvals = recurEval  parent  rootplayer   BLCs\n");
	for (int i = 0; i < (int)tree.nodes.size(); i++) {
		fprintfBriefPPNode(fp, tree.nodes.at(i), i);
	}
	for (int i = 0; i < (int)tree.nodes.size(); i++) {
		fprintfPPNode(fp, tree.nodes.at(i), i);
	}
}


// Prints one line per node. Does not print BLCs
static void fprintfTreeSummary(FILE *fp, const PPTree & tree) {
	fprintf(fp, "Node   seed                                 trueEval  estEval  recurSum/nEvals = recurEval  parent  rootplayer   BLCs\n");
	for (int i = 0; i < (int)tree.nodes.size(); i++) {
		fprintfBriefPPNode(fp, tree.nodes.at(i), i);
	}
}




/*********************************************************************************************************/
/************************************** general tree operations ******************************************/
/*********************************************************************************************************/





static int makeNumberOfMoves(int parentNmoves, double nMovesScale, double nMovesStdDev, std::default_random_engine & prng) {
	std::normal_distribution<double> normDist(0.5*(parentNmoves + nMovesScale), nMovesStdDev);
	double x = normDist(prng);
	int xx = (int)floor(x + 0.5);
	xx = std::max(xx, 1);
	xx = std::min(xx, variabletreeConfig.nMovesMaximum);
	return xx;
}


static std::vector<double> makeTrueEvals(double parentTrueEval, int nMoves, std::default_random_engine & prng) {
	std::vector<double> trueEvals(nMoves);
	double spacingRate = sqrt(nMoves) / currenttreeConfig.trueEvalspread;
	std::exponential_distribution<double> rexp(spacingRate);
	trueEvals[0] = 0.0;
	for (int mov = 1; mov < nMoves; mov++) {
		trueEvals[mov] = trueEvals[mov - 1] + rexp(prng);
	}
	for (int mov = 0; mov < nMoves; mov++) {
		trueEvals[mov] = std::min(trueEvals[mov], MAXSPREAD);
	}
	double w = 1.0 - parentTrueEval;
	for (int mov = 0; mov < nMoves; mov++) {
		trueEvals[mov] = w / (w + exp(-trueEvals[mov]) * (1.0 - w));
		trueEvals[mov] = std::max(trueEvals[mov], MINEVAL);
		trueEvals[mov] = std::min(trueEvals[mov], MAXEVAL);
	}	
	return trueEvals;
}



static double estEvalNoise(std::default_random_engine & prng) {
	std::student_t_distribution<double> tdist(currenttreeConfig.estEvaldf);
	return currenttreeConfig.estEvalscaleCh * tdist(prng);
}



static double estEval(double trueEval, double parentNoise, double estNoise) {
	double parcorr = currenttreeConfig.estEvalscalePar;
	double y = estNoise - parcorr * parentNoise;
	y /= sqrt(1.0 + parcorr * parcorr);
	double x = trueEval;
	double z = x / (x + exp(y)*(1.0 - x));
	assert(!std::isnan(z));
	z = std::max(z, MINEVAL);
	z = std::min(z, MAXEVAL);
	return z;
}



static double policyEval(double trueEval, double estNoise, std::default_random_engine & prng) {
	std::student_t_distribution<double> tdist(currenttreeConfig.policyEvaldf);
	double y = currenttreeConfig.policyEvalscaleCh * tdist(prng);
	y += currenttreeConfig.policyEvalscaleEE * estNoise;
	double x = trueEval;
	double z = x / (x + exp(y)*(1.0 - x));
	assert(!std::isnan(z));
	z = std::max(z, MINEVAL);
	z = std::min(z, MAXEVAL);	
	return z;
}


// makeBuds() makes Buds for a new Node. It actually makes Leaves (they have evaluations) but they
// are marked isLeaf=false, so treated as Buds initially.
// makeBuds() seeds the PRNG from the move sequence. The number of moves, the estimated eval,
// and the policy vector are then sampled from the PRNG.
// Some information is put into BudLeafChild's and then transferred to a Node
// if the move is expanded, which may look odd or inefficient. This is so that all the 
// random quantities can be added to a Node at once. 
static std::vector<BudLeafChild> makeBuds(int parentNmoves, PPNode & ppnode) {
	std::default_random_engine prng;
	std::vector<int> movesplus = ppnode.moveSeq;
	movesplus.push_back(currenttreeConfig.BLsSeed);
	std::seed_seq seq(movesplus.begin(), movesplus.end());
	prng.seed(seq);
	int nMoves = makeNumberOfMoves(parentNmoves, currenttreeConfig.nMovesScale, currenttreeConfig.nMovesStdDev, prng);
	std::vector<BudLeafChild> blcs;
	std::vector<double> policyEvals;
	std::vector<double> trueEvals = makeTrueEvals(ppnode.trueEval, nMoves, prng);
	for (int mov = 0; mov < nMoves; mov++) {
		BudLeafChild blc;
		blc.isChild = false;
		blc.isLeaf = false; // initially a Bud
		blc.childIdx = -1;
		blc.trueEval = trueEvals[mov];
		blc.estNoise = estEvalNoise(prng);
		blc.estEval = estEval(blc.trueEval, ppnode.estNoise, blc.estNoise);
		blc.policy = -1.0;
		blc.recurPolicy = -1.0; // -1 meaning 'not set' 
		blcs.push_back(blc);
		policyEvals.push_back(policyEval(blc.trueEval, blc.estNoise, prng));
	}
	double softmaxTotal = 0.0;
	double smS = currenttreeConfig.policySoftmaxScale;
	for (int mov = 0; mov < nMoves; mov++) {
		softmaxTotal += exp(-smS * policyEvals[mov]);
	}
	for (int mov = 0; mov < nMoves; mov++) {
		blcs[mov].policy = exp(-smS * policyEvals[mov]) / softmaxTotal;
	}
	return blcs;
}




/*********************************************************************************************************/
/*********************************** checking/debugging *******************************************/
/*********************************************************************************************************/

static void recursiveCheckEvals(PPTree & tree, int n) {
	if (nChildren(tree.nodes[n]) == 0) {
		assert(tree.nodes[n].recurMaxNeg == tree.nodes[n].estEval);
		assert(tree.nodes[n].recurSum == tree.nodes[n].estEval);
		assert(tree.nodes[n].nEvals == 1);
	} else {
		double max = 0.0;
		double sum = 0.0;
		int nEvals = 0;
		for (BudLeafChild leaf : tree.nodes[n].BLCs) {
			if (leaf.isChild) {
				int chi = leaf.childIdx;
				recursiveCheckEvals(tree, chi);
				max = std::max(max, 1.0 - tree.nodes[chi].recurMaxNeg);
				sum += (tree.nodes[chi].nEvals - tree.nodes[chi].recurSum); 
				nEvals += tree.nodes[chi].nEvals;
			}
		}
		
		//assert(tree.nodes[n].recurMaxNeg == max);// TODO
		double shouldbe = tree.nodes[n].estEval + sum;
		if (fabs((tree.nodes[n].recurSum - shouldbe)/shouldbe) >= 1e-10) {
			printf("%f %f\n", tree.nodes[n].recurSum, shouldbe);
			getchar();
		}
 		assert(fabs((tree.nodes[n].recurSum - shouldbe)/shouldbe) < 1e-10);
		assert(tree.nodes[n].nEvals == 1 + nEvals);
	}
}




/*********************************************************************************************************/
/***********************************   for small test trees    *******************************************/
/*********************************************************************************************************/


static void allBudsToChildren(PPTree & tree, int idx) {
	LeafIndex lfi;
	lfi.idx = idx;
	for (int mov = 0; mov < (int)tree.nodes[idx].BLCs.size(); mov++) {
		lfi.mov = mov;
		budToChild(tree, lfi);
	}
}










/***********************************************************************************************/
/***********************************************************************************************/
/***********************************************************************************************/
/***********************************************************************************************/
/***********************************************************************************************/
/***********************************************************************************************/
/***********************************       extern         **************************************/
/***********************************************************************************************/
/***********************************************************************************************/
/***********************************************************************************************/
/***********************************************************************************************/
/***********************************************************************************************/
/***********************************************************************************************/
/***********************************************************************************************/
/***********************************************************************************************/


void budToLeaf(PPTree & tree, LeafIndex lfi) {
	int idx = lfi.idx;
	int mov = lfi.mov;
	tree.nodes[idx].BLCs[mov].isLeaf = true;
}


void leafToChild(PPTree & tree, LeafIndex lfi) {
	PPNode child;
	int idx = lfi.idx;
	int mov = lfi.mov;
	child.trueEval = tree.nodes[idx].BLCs.at(mov).trueEval;
	child.estNoise = tree.nodes[idx].BLCs.at(mov).estNoise;
	child.estEval = tree.nodes[idx].BLCs.at(mov).estEval;
	child.nEvals = 1;
	child.realProb = tree.nodes[idx].realProb * tree.nodes[idx].BLCs[mov].policy;
	child.moveSeq = tree.nodes[idx].moveSeq;
	child.moveSeq.push_back(mov);
	child.BLCs = makeBuds((int)tree.nodes[idx].BLCs.size(), child);
	child.parent = idx;
	child.rootplayer = !tree.nodes[idx].rootplayer;
	tree.nodes[idx].BLCs[mov].isChild = true;
	tree.nodes[idx].BLCs[mov].childIdx = (int)tree.nodes.size();
	child.recurSum = child.estEval;
	child.recurMaxNeg = child.estEval;
	tree.nodes.push_back(child);
}


void budToChild(PPTree & tree, LeafIndex lfi) {
	budToLeaf(tree, lfi);
	leafToChild(tree, lfi);
}



bool isBud(BudLeafChild & blc) {
	return !blc.isLeaf && !blc.isChild;
}

bool isLeaf(BudLeafChild & blc) {
	return blc.isLeaf && !blc.isChild;
}

bool isChild(BudLeafChild & blc) {
	return blc.isChild;
}

bool isBudOrChild(BudLeafChild & blc) {
	return !blc.isLeaf || blc.isChild;
}

bool isLeafOrChild(BudLeafChild & blc) {
	return blc.isLeaf;
}



void makeRoot(PPTree & tree) {
	uint32_t seed = currenttreeConfig.rootSeed;
	std::default_random_engine prng;
	prng.seed(seed);
	PPNode root;
	root.trueEval = currenttreeConfig.trueRootEval;
	root.estNoise = estEvalNoise(prng);
	root.estEval = estEval(root.trueEval, 0.0, root.estNoise);
	root.parent = -1;
	root.rootplayer = true;
	root.realProb = 1.0;
	root.nEvals = 1; // as if root's parent made the root
	root.moveSeq.push_back(0);
	root.BLCs = makeBuds((int)currenttreeConfig.nMovesOfRootsParent, root);
	root.recurSum = root.estEval;
	root.recurMaxNeg = root.estEval;
	tree.nodes.push_back(root);
}


void  doEvalsBackToRoot(PPTree & tree, int chi) {
	int here = chi;
	int parent = tree.nodes[here].parent;
	double evalToAdd = tree.nodes[here].recurSum;
	bool chiside = false;
	while (parent >= 0) {
		tree.nodes[parent].nEvals += 1;
		tree.nodes[parent].recurSum += chiside ? evalToAdd : (1.0 - evalToAdd);
		here = parent;
		parent = tree.nodes[here].parent;
		chiside = !chiside;
	}
}




int collectEvalsAndNumEvals(const PPTree & tree, int idx, 
	        std::vector<double> & staticPol, std::vector<double> & evals, std::vector<int> & numEvals) {
	const PPNode & node = tree.nodes[idx];
	int nMoves = (int)node.BLCs.size();
	int total_nm = 0;
	for (int mov = 0; mov < nMoves; mov++) {
		BudLeafChild blc = node.BLCs[mov];
		int nParentEvals = node.nEvals;
		staticPol[mov] = node.BLCs[mov].policy;
		if (blc.isChild) {
			const PPNode & child = tree.nodes[blc.childIdx];
			evals[mov] = 1.0 - child.recurSum / child.nEvals;
			numEvals[mov] = child.nEvals;
		}
		else {
			assert(!blc.isLeaf); // with 'plain' searches, bud goes straight to child
			evals[mov] = 0.0;
			numEvals[mov] = 0;
		}
		total_nm += numEvals[mov];
	}
	assert(node.nEvals == 1 + total_nm);
	return  total_nm;
}


// Backup eval sums, counts, minimax
void subtreeBackupEvalsRPT(PPTree & tree, int n) {
	if (nChildren(tree.nodes[n]) == 0) {
		tree.nodes[n].recurMaxNeg = tree.nodes[n].estEval;
		tree.nodes[n].recurSum = tree.nodes[n].estEval;
		tree.nodes[n].nEvals = 1;
	}
	else {
		double max = 0.0;
		double sum = 0.0;
		int nEvals = 0;
		for (BudLeafChild leaf : tree.nodes[n].BLCs) {
			if (leaf.isChild) {
				int chi = leaf.childIdx;
				subtreeBackupEvalsRPT(tree, chi);
				max = std::max(max, 1.0 - tree.nodes[chi].recurMaxNeg);
				sum += (tree.nodes[chi].nEvals - tree.nodes[chi].recurSum);
				nEvals += tree.nodes[chi].nEvals;
			}
		}
		assert(!std::isnan(sum));
		tree.nodes[n].recurMaxNeg = max;
		tree.nodes[n].recurSum = tree.nodes[n].estEval + sum;
		tree.nodes[n].nEvals = 1 + nEvals;
	}
}




void fprintfResults(FILE *fp, std::vector<const PPNode *> ply1nodes) {
	for (int i = 0; i < (int)ply1nodes.size(); i++) {
		fprintf(fp, "%d %d %6.4f\n", ply1nodes[i]->moveSeq[1], ply1nodes[i]->nEvals, ply1nodes[i]->realProb);
	}
	fprintf(fp, "\n");
}


void writeTree(const char * fpath, const PPTree & tree) {
	if (variabletreeConfig.debugLevel > 0) {
		TreeStatistics ts = collectTreeStatistics(tree);
		FILE *fp = openFileForWriting(fpath);
		fprintfSimtreeConfig(fp);
		fprintfTreeStats(fp, ts);
		if (variabletreeConfig.debugLevel == 1) {
			fprintfTreeSummary(fp, tree);
		}
		if (variabletreeConfig.debugLevel >= 2) {
			fprintfPPTree(fp, tree);
		}
		fclose(fp);
	}
}



std::vector<const PPNode *> get1plyNodes(const PPTree & tree)
{
	std::vector<const PPNode *> ply1nodes;
	for (int i = 0; i < (int)tree.nodes.size(); i++) {
		if (tree.nodes[i].moveSeq.size() == 2) {
			ply1nodes.push_back(&tree.nodes[i]);
		}
	}
	return ply1nodes;
}




void fscanfVariableSimtreeConfig(FILE *fp) {
	fscanf(fp, " Simulation parameters");
	char line[300];
	readNextNonemptyline(fp, line);
	variabletreeConfig.nMovesOfRootsParentSet = gatherValues("nMovesOfRootsParent", line);
	readNextNonemptyline(fp, line);
	variabletreeConfig.nMovesScaleSet = gatherValues("nMovesScale", line);
	readNextNonemptyline(fp, line);
	variabletreeConfig.nMovesStdDevSet = gatherValues("nMovesStdDev", line);
	assert(1 == fscanf(fp, " nMovesMaximum %d", &variabletreeConfig.nMovesMaximum));

	readNextNonemptyline(fp, line);
	variabletreeConfig.trueRootEvalSet = gatherValues("trueRootEval", line);
	readNextNonemptyline(fp, line);
	variabletreeConfig.trueEvalspreadSet = gatherValues("trueEvalspread", line);

	readNextNonemptyline(fp, line);
	variabletreeConfig.estEvaldfSet = gatherValues("estEvaldf", line);
	readNextNonemptyline(fp, line);
	variabletreeConfig.estEvalscaleParSet = gatherValues("estEvalscalePar", line);
	readNextNonemptyline(fp, line);
	variabletreeConfig.estEvalscaleChSet = gatherValues("estEvalscaleCh", line);
	readNextNonemptyline(fp, line);

	variabletreeConfig.policyEvaldfSet = gatherValues("policyEvaldf", line);
	readNextNonemptyline(fp, line);
	variabletreeConfig.policyEvalscaleChSet = gatherValues("policyEvalscaleCh", line);
	readNextNonemptyline(fp, line);
	variabletreeConfig.policyEvalscaleEESet = gatherValues("policyEvalscaleEE", line);
	readNextNonemptyline(fp, line);
	variabletreeConfig.policySoftmaxScaleSet = gatherValues("policySoftmaxScale", line);

	assert(1 == fscanf(fp, " rootSeed %d", &variabletreeConfig.rootSeed));
	assert(1 == fscanf(fp, " BLsSeed %d", &variabletreeConfig.BLsSeed));
	assert(2 == fscanf(fp, " treeIndicies %d %d", &variabletreeConfig.firstTree, &variabletreeConfig.endtrees));

	assert(1 == fscanf(fp, " debugLevel %d", &variabletreeConfig.debugLevel));
}



void fprintfVariableSimtreeConfig(FILE *fp) {
	fprintf(fp, "    Simulation parameters\n");
	fprintfVectorValues(fp, "nMovesOfRootsParent", variabletreeConfig.nMovesOfRootsParentSet);
	fprintfVectorValues(fp, "nMovesScale", variabletreeConfig.nMovesScaleSet);
	fprintfVectorValues(fp, "nMovesStdDev", variabletreeConfig.nMovesStdDevSet);
	fprintf(fp, "    nMovesMaximum %d\n", variabletreeConfig.nMovesMaximum);

	fprintfVectorValues(fp, "trueRootEval", variabletreeConfig.trueRootEvalSet);
	fprintfVectorValues(fp, "trueEvalspreadSet", variabletreeConfig.trueEvalspreadSet);

	fprintfVectorValues(fp, "estEvaldfSet", variabletreeConfig.estEvaldfSet);
	fprintfVectorValues(fp, "estEvalscaleParSet", variabletreeConfig.estEvalscaleParSet);
	fprintfVectorValues(fp, "estEvalscaleChSet", variabletreeConfig.estEvalscaleChSet);

	fprintfVectorValues(fp, "policyEvaldfSet", variabletreeConfig.policyEvaldfSet);
	fprintfVectorValues(fp, "policyEvalscaleChSet", variabletreeConfig.policyEvalscaleChSet);
	fprintfVectorValues(fp, "policyEvalscaleEESet", variabletreeConfig.policyEvalscaleEESet);
	fprintfVectorValues(fp, "policySoftmaxScaleSet", variabletreeConfig.policySoftmaxScaleSet);

	fprintf(fp, "    rootSeed %d\n", variabletreeConfig.rootSeed);
	fprintf(fp, "    BLsSeed %d\n", variabletreeConfig.BLsSeed);

	fprintf(fp, "    treeIndicies %d %d\n\n", variabletreeConfig.firstTree, variabletreeConfig.endtrees);
}



void fprintfSimtreeConfig(FILE *fp) {
	fprintf(fp, "    Simulation parameters\n");
	fprintf(fp, "    nMovesOfRootsParent %f\n", currenttreeConfig.nMovesOfRootsParent);
	fprintf(fp, "    nMovesScale %f\n", currenttreeConfig.nMovesScale);
	fprintf(fp, "    nMovesStdDev %f\n", currenttreeConfig.nMovesStdDev);

	fprintf(fp, "    trueRootEval %f\n", currenttreeConfig.trueRootEval);
	fprintf(fp, "    trueEvalspread %f\n", currenttreeConfig.trueEvalspread);

	fprintf(fp, "    estEvaldf %f\n", currenttreeConfig.estEvaldf);
	fprintf(fp, "    estEvalscalePar %f\n", currenttreeConfig.estEvalscalePar);
	fprintf(fp, "    estEvalscaleCh %f\n", currenttreeConfig.estEvalscaleCh);

	fprintf(fp, "    policyEvaldf %f\n", currenttreeConfig.policyEvaldf);
	fprintf(fp, "    policyEvalscaleCh %f\n", currenttreeConfig.policyEvalscaleCh);
	fprintf(fp, "    policyEvalscaleEE %f\n", currenttreeConfig.policyEvalscaleEE);
	fprintf(fp, "    policySoftmaxScale %f\n", currenttreeConfig.policySoftmaxScale);

	fprintf(fp, "    treeIndex %d\n", currenttreeConfig.treeIndex);
	fprintf(fp, "    rootSeed %d\n", currenttreeConfig.rootSeed);
	fprintf(fp, "    BLsSeed %d\n", currenttreeConfig.BLsSeed);
}


SearchResults searchResultsFromTree(const PPTree & tree, bool useRPs) {
	SearchResults results;
	const PPNode & root = tree.nodes[0];
	int n1ply = (int)root.BLCs.size();
	results.trueRootEval = root.trueEval;
	results.origPolicy = std::vector<double>(n1ply);
	results.recurPolicy = std::vector<double>(n1ply);
	results.true1plyEvals = std::vector<double>(n1ply);
	for (int mov = 0; mov < n1ply; mov++) {
		results.origPolicy[mov] = root.BLCs[mov].policy;
		results.recurPolicy[mov] = root.BLCs[mov].recurPolicy;
		results.true1plyEvals[mov] = root.BLCs[mov].trueEval;
	}
	std::vector<const PPNode *> ply1nodes = get1plyNodes(tree);
	if (useRPs) {
		std::sort(ply1nodes.begin(), ply1nodes.end(), [](const PPNode * a, const PPNode * b) {
			return a->realProb > b->realProb;
		});
	}
	else {
		std::sort(ply1nodes.begin(), ply1nodes.end(), [](const PPNode * a, const PPNode * b) {
			return a->nEvals > b->nEvals;
		});
	}
	int nBest = 1;
	while (nBest < ply1nodes.size() && ply1nodes[nBest]->nEvals == ply1nodes[0]->nEvals) {
		nBest++;
	}
	results.choice = ply1nodes[nBest / 2]->moveSeq[1];
	results.evalmean = root.recurSum / root.nEvals;
	results.evalminimax = root.recurMaxNeg;

	return results;
}






void makeTestTree2Ply(PPTree & tree) {
	makeRoot(tree);
	// expand root's buds to ply1 Nodes
	allBudsToChildren(tree, 0);
	// expand ply1 buds to ply2 Nodes
	for (BudLeafChild blc1 : tree.nodes[0].BLCs) {
		assert(blc1.isChild);
		allBudsToChildren(tree, blc1.childIdx);
	}
}



void makeTestTree3Ply(PPTree & tree) {
	makeTestTree2Ply(tree);
	// expand ply2 buds to ply3 Nodes
	for (BudLeafChild blc1 : tree.nodes[0].BLCs) {
		for (BudLeafChild blc2 : tree.nodes[blc1.childIdx].BLCs) {
			assert(blc2.isChild);
			allBudsToChildren(tree, blc2.childIdx);
		}
	}
}


void makeTestTree4Ply(PPTree & tree) {
	makeTestTree3Ply(tree);
	// expand ply3 buds to ply4 Nodes
	for (BudLeafChild blc1 : tree.nodes[0].BLCs) {
		for (BudLeafChild blc2 : tree.nodes[blc1.childIdx].BLCs) {
			for (BudLeafChild blc3 : tree.nodes[blc2.childIdx].BLCs) {
				assert(blc3.isChild);
				allBudsToChildren(tree, blc3.childIdx);
			}
		}
	}
}



