#pragma once

#include <vector>



// Evals are expected game scores in [0,1]. They are from the point of view of the player 
// to move. So it's 'negamin' or 'maxneg':
//     parent eval  =  1 - min(child evals)  =  max((1 - child eval)s)
// trueEvals are ordered, so that move 0 is the minimum. The PV is move sequence 0, 0, 0, 0, ...

// To add noise to an evaluation p in [0,1], I map to (-oo,+oo) via inverse 
// logistic cdf. x = -log(1/p-1). Then I add symmetric noise
// to x and map back. I call the space where x lives "noiseSLI space".
// It is a space where the noise is symmetric and location independent.
// In noiseSLI space, the noise is a scaled Student  t-distribution.

// Each Node seeds the PRNG with the move sequence from root, so 
// that all searches explore the same virtual tree.



#define CUDAMIMIC_MAX_MOVES 100


typedef struct {
	bool isLeaf;
	bool isChild;
	int childIdx;
	double trueEval;
	double estNoise;  // In noiseSLI space
	double estEval;   
	double policy;      // = initial realisation probabilities
	double recurPolicy; // = recursive policy = conditional realisation probability
} BudLeafChild;
//  isLeaf = false, isChild = false (Bud)  OR  
//  isLeaf = true,  isChild = false (Leaf)  OR  
//  isLeaf = true,  isChild = true  (points to child Node)


typedef struct {
	std::vector<int> moveSeq;
	double trueEval;  // in [0,1]
	double estNoise;  // In noiseSLI space
	double estEval;   // in [0,1]
	double recurSum;
	double recurMaxNeg;
	int nEvals;
	double realProb; // = (unconditional) realisation probabilities
	std::vector<BudLeafChild> BLCs;
	int parent;
	bool rootplayer;
} PPNode;


// The simulation parameters for a single tree
typedef struct {
	double nMovesOfRootsParent; // For number of moves from root (only).
	double nMovesScale;         // For number of moves from any node
	double nMovesStdDev;        // ditto		

	double trueRootEval;        // The true eval of the root in [0,1]
	double trueEvalspread;      // controls how much worse true evals get for later moves.
						        // Successive differences noiseSLI space between true evals are iid 
						        // exponentials with rate = sqrt(nMoves) / treeConfig.trueEvalspread

	double estEvaldf;	        // control noise in evals (shape)
	double estEvalscalePar;     // noise inherited from parent
	double estEvalscaleCh;      // independent noise for each child

	double policyEvaldf;        // control noise in policy (shape)
	double policyEvalscaleCh;   // independent noise for each child
	double policyEvalscaleEE;   // noise transferred from eval noise
	double policySoftmaxScale;  // controls how concentrated the policy vector is

	int treeIndex;
	int rootSeed;               // For root est eval
	int BLsSeed;                // for making buds/leaves of a new node

} SimTreeConfig;


// The simulation parameters for a set of trees.
// Each vector<double> contains one or more values which are chosen
// uniformly and independently of one another.
// For each tree, an index is added to the the seeds to make 
// corresponding values in SimTreeConfig.
typedef struct {
	std::vector<double> nMovesOfRootsParentSet;
	std::vector<double> nMovesScaleSet;
	std::vector<double> nMovesStdDevSet;	 
	int nMovesMaximum;      	

	std::vector<double> trueRootEvalSet;
	std::vector<double> trueEvalspreadSet;
	
	std::vector<double> estEvaldfSet;
	std::vector<double> estEvalscaleParSet;
	std::vector<double> estEvalscaleChSet;

	std::vector<double> policyEvaldfSet;
	std::vector<double> policyEvalscaleChSet;
	std::vector<double> policyEvalscaleEESet;
	std::vector<double> policySoftmaxScaleSet;

	int rootSeed;   // For choosing among the sets above, and for root est eval.
	int BLsSeed;    // For making buds/leaves of a new node.

	int firstTree;  // range of trees to use
	int endtrees;

	int debugLevel;             // controls verbosity of tree output, etc. 0=off, 1=summary, 2=full
} VariableSimTreeConfig;




typedef struct {
	std::vector<PPNode> nodes;

} PPTree;



typedef struct {
	int idx;
	int mov;
} LeafIndex;


typedef struct {
	double trueRootEval;
	std::vector<double> true1plyEvals;
	std::vector<double> origPolicy;
	std::vector<double> recurPolicy;
	int choice;
	double evalmean;
	double evalminimax;
} SearchResults;




constexpr auto MAXEVAL = 0.99999;
constexpr auto MINEVAL = 0.00001;
constexpr auto MAXSPREAD = 50.0;

extern VariableSimTreeConfig variabletreeConfig;
extern SimTreeConfig currenttreeConfig;
extern FILE *debugInfo_fp;


// Mark BLC as leaf, set parent nEvals=1.
void budToLeaf(PPTree & tree, LeafIndex lfi);

// Expansion. Create a new Node with its own Buds. True and estimated evals are
// copied from parent's BLC, other things initialised. Parent's BLC marked as child
// and points to new Node.
void leafToChild(PPTree & tree, LeafIndex lfi);

// Calls budToLeaf(), leafToChild(), for modelling case where eval and policy of child found together (as LC0).
void budToChild(PPTree & tree, LeafIndex lfi);

bool isBud(BudLeafChild & blc);
bool isLeaf(BudLeafChild & blc);
bool isChild(BudLeafChild & blc);
bool isBudOrChild(BudLeafChild & blc);
bool isLeafOrChild(BudLeafChild & blc);

void makeRoot(PPTree & tree);


// updates nEvals and recurSum up path from chi to root.
void  doEvalsBackToRoot(PPTree & tree, int chi);

// fills staticPol with p_m, evals[] with v_m, numEvals[] with n_m, u
// using budEval for v_m when n_m is zero.
// Returns total evals below node (not including the one at node)
int collectEvalsAndNumEvals(const PPTree & tree, int idx,
            std::vector<double> & staticPol, std::vector<double> & evals, std::vector<int> & numEvals);

// For calculating results after search.
std::vector<const PPNode *> get1plyNodes(const PPTree & tree);

// Updates fields recurMaxNeg, nEvals, recurSum from tips for all Nodes
// below n. Call with n=0 for whole tree.
void subtreeBackupEvalsRPT(PPTree & tree, int n);

// Write whole tree (node array) to file
void writeTree(const char * fpath, const PPTree & tree);


void fprintfResults(FILE *fp, std::vector<const PPNode *> ply1nodes);

// Reading and writing config
void fscanfVariableSimtreeConfig(FILE *fp);
void fprintfVariableSimtreeConfig(FILE *fp);
void fprintfSimtreeConfig(FILE *fp);

// For making (small) test trees. For debugging, checking
void makeTestTree2Ply(PPTree & tree);
void makeTestTree3Ply(PPTree & tree);
void makeTestTree4Ply(PPTree & tree);



SearchResults searchResultsFromTree(const PPTree & tree, bool useRPs);
